var _a;
import { LocalNode } from "cojson";
import { Context } from "effect";
import { Group, CoValueBase, MembersSym, Profile, Ref, SchemaInit, inspect, subscriptionsScopes, } from "../internal.js";
export class Account extends CoValueBase {
    get _schema() {
        return this.constructor._schema;
    }
    get _owner() {
        return this;
    }
    get _loadedAs() {
        return this.isMe
            ? this
            : _a.fromNode(this._raw.core.node);
    }
    get _refs() {
        const profileID = this._raw.get("profile");
        const rootID = this._raw.get("root");
        return {
            profile: profileID &&
                new Ref(profileID, this._loadedAs, this._schema.profile),
            root: rootID &&
                new Ref(rootID, this._loadedAs, this._schema.root),
        };
    }
    constructor(_init, options) {
        super();
        this.acceptInvite = (async (valueID, inviteSecret, coValueClass) => {
            if (!this.isMe) {
                throw new Error("Only a controlled account can accept invites");
            }
            await this._raw.acceptInvite(valueID, inviteSecret);
            return coValueClass.load(valueID, {
                as: this,
            });
        });
        if (!("fromRaw" in options)) {
            throw new Error("Can only construct account from raw or with .create()");
        }
        this.isMe = options.fromRaw.id == options.fromRaw.core.node.account.id;
        Object.defineProperties(this, {
            id: {
                value: options.fromRaw.id,
                enumerable: false,
            },
            _raw: { value: options.fromRaw, enumerable: false },
        });
        if (this.isMe) {
            this.sessionID =
                options.fromRaw.core.node.currentSessionID;
        }
        return new Proxy(this, AccountAndGroupProxyHandler);
    }
    myRole() {
        if (this.isMe) {
            return "admin";
        }
    }
    static async create(options) {
        const { node } = await LocalNode.withNewlyCreatedAccount({
            ...options,
            migration: async (rawAccount, _node, creationProps) => {
                const account = new this(undefined, {
                    fromRaw: rawAccount,
                });
                await account.migrate?.(creationProps);
            },
        });
        return this.fromNode(node);
    }
    static async become(options) {
        const node = await LocalNode.withLoadedAccount({
            accountID: options.accountID,
            accountSecret: options.accountSecret,
            sessionID: options.sessionID,
            peersToLoadFrom: options.peersToLoadFrom,
            migration: async (rawAccount, _node, creationProps) => {
                const account = new this(undefined, {
                    fromRaw: rawAccount,
                });
                await account.migrate?.(creationProps);
            },
        });
        return this.fromNode(node);
    }
    static fromNode(node) {
        return new this(undefined, {
            fromRaw: node.account,
        });
    }
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    migrate(creationProps) {
        if (creationProps) {
            const profileGroup = new Group({ owner: this });
            profileGroup.addMember("everyone", "reader");
            this.profile = new Profile({ name: creationProps.name }, { owner: profileGroup });
        }
    }
}
_a = Account;
(() => {
    _a._schema = {
        profile: () => Profile,
        root: "json",
    };
})();
export const AccountAndGroupProxyHandler = {
    get(target, key, receiver) {
        if (key === "profile") {
            const ref = target._refs.profile;
            return ref ? ref.accessFrom(receiver) : undefined;
        }
        else if (key === "root") {
            const ref = target._refs.root;
            return ref ? ref.accessFrom(receiver) : undefined;
        }
        else {
            return Reflect.get(target, key, receiver);
        }
    },
    set(target, key, value, receiver) {
        var _b;
        if ((key === "profile" || key === "root" || key === MembersSym) &&
            typeof value === "object" &&
            SchemaInit in value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[key] =
                value[SchemaInit];
            return true;
        }
        else if (key === "profile") {
            if (value) {
                target._raw.set("profile", value.id, "trusting");
            }
            subscriptionsScopes.get(receiver)?.onRefAccessedOrSet(value.id);
            return true;
        }
        else if (key === "root") {
            if (value) {
                target._raw.set("root", value.id);
            }
            subscriptionsScopes.get(receiver)?.onRefAccessedOrSet(value.id);
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, descriptor) {
        var _b;
        if ((key === "profile" || key === "root" || key === MembersSym) &&
            typeof descriptor.value === "object" &&
            SchemaInit in descriptor.value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[key] =
                descriptor.value[SchemaInit];
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, descriptor);
        }
    },
};
export class AccountCtx extends Context.Tag("Account")() {
}
export function isControlledAccount(account) {
    return account.isMe;
}
//# sourceMappingURL=account.js.map