var _a;
import { encodeSync, decodeSync } from "@effect/schema/Schema";
import { Account, CoValueBase, Ref, SchemaInit, inspect, makeRefs, subscriptionsScopes, ItemsSym, InitValues, isRefEncoded, } from "../internal.js";
export class CoMap extends CoValueBase {
    get _schema() {
        return this.constructor._schema;
    }
    get _refs() {
        return makeRefs((key) => this._raw.get(key), () => {
            const keys = this._raw.keys().filter((key) => {
                const schema = this._schema[key] ||
                    this._schema[ItemsSym];
                return schema && schema !== "json" && isRefEncoded(schema);
            });
            return keys;
        }, this._loadedAs, (key) => (this._schema[key] ||
            this._schema[ItemsSym]));
    }
    get _edits() {
        return new Proxy(this, {
            get(target, key) {
                const rawEdit = target._raw.lastEditAt(key);
                if (!rawEdit)
                    return undefined;
                const descriptor = target._schema[key];
                return {
                    value: descriptor === "json"
                        ? rawEdit.value
                        : "encoded" in descriptor
                            ? decodeSync(descriptor.encoded)(rawEdit.value)
                            : new Ref(rawEdit.value, target._loadedAs, descriptor).accessFrom(target),
                    ref: descriptor !== "json" && isRefEncoded(descriptor)
                        ? new Ref(rawEdit.value, target._loadedAs, descriptor)
                        : undefined,
                    by: rawEdit.by &&
                        new Ref(rawEdit.by, target._loadedAs, Account).accessFrom(target),
                    madeAt: rawEdit.at,
                };
            },
        });
    }
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    constructor(init, options) {
        super();
        if (init && "owner" in options) {
            this[InitValues] = {
                init,
                owner: options.owner,
            };
        }
        else if ("fromRaw" in options) {
            Object.defineProperties(this, {
                id: {
                    value: options.fromRaw.id,
                    enumerable: false,
                },
                _raw: { value: options.fromRaw, enumerable: false },
            });
        }
        else {
            throw new Error("Invalid CoMap constructor arguments");
        }
        return new Proxy(this, CoMapProxyHandler);
    }
    toJSON() {
        const jsonedFields = this._raw.keys().map((key) => {
            const tKey = key;
            const descriptor = (this._schema[tKey] ||
                this._schema[ItemsSym]);
            if (descriptor == "json" || "encode" in descriptor) {
                return [key, this._raw.get(key)];
            }
            else if (isRefEncoded(descriptor)) {
                const jsonedRef = this[tKey]?.toJSON();
                return [key, jsonedRef];
            }
            else {
                return [key, undefined];
            }
        });
        return {
            id: this.id,
            _type: this._type,
            ...Object.fromEntries(jsonedFields),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    rawFromInit(init, owner) {
        const rawOwner = owner._raw;
        const rawInit = {};
        if (init)
            for (const key of Object.keys(init)) {
                const initValue = init[key];
                const descriptor = (this._schema[key] || this._schema[ItemsSym]);
                if (descriptor === "json") {
                    rawInit[key] = initValue;
                }
                else if (isRefEncoded(descriptor)) {
                    if (initValue) {
                        rawInit[key] = initValue.id;
                    }
                }
                else if ("encoded" in descriptor) {
                    rawInit[key] = encodeSync(descriptor.encoded)(initValue);
                }
            }
        return rawOwner.createMap(rawInit);
    }
    static Record(value) {
        var _b;
        // eslint-disable-next-line @typescript-eslint/no-unsafe-declaration-merging
        class RecordLikeCoMap extends _a {
            constructor() {
                super(...arguments);
                this[_b] = value;
            }
        }
        _b = ItemsSym;
        return RecordLikeCoMap;
    }
}
_a = CoMap;
(() => {
    _a.prototype._type = "CoMap";
})();
function tryInit(map) {
    if (map[InitValues] &&
        (map._schema[ItemsSym] ||
            Object.keys(map[InitValues].init).every((key) => map._schema[key]))) {
        const raw = map.rawFromInit(map[InitValues].init, map[InitValues].owner);
        Object.defineProperties(map, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        delete map[InitValues];
    }
}
// TODO: cache handlers per descriptor for performance?
const CoMapProxyHandler = {
    get(target, key, receiver) {
        if (key === "_schema") {
            return Reflect.get(target, key);
        }
        else if (key in target) {
            return Reflect.get(target, key, receiver);
        }
        else {
            const descriptor = (target._schema[key] ||
                target._schema[ItemsSym]);
            if (descriptor && typeof key === "string") {
                const raw = target._raw.get(key);
                if (descriptor === "json") {
                    return raw;
                }
                else if ("encoded" in descriptor) {
                    return raw === undefined
                        ? undefined
                        : decodeSync(descriptor.encoded)(raw);
                }
                else if (isRefEncoded(descriptor)) {
                    return raw === undefined
                        ? undefined
                        : new Ref(raw, target._loadedAs, descriptor).accessFrom(receiver);
                }
            }
            else {
                return undefined;
            }
        }
    },
    set(target, key, value, receiver) {
        var _b;
        if ((typeof key === "string" || ItemsSym) &&
            typeof value === "object" &&
            SchemaInit in value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[key] =
                value[SchemaInit];
            tryInit(target);
            return true;
        }
        const descriptor = (target._schema[key] ||
            target._schema[ItemsSym]);
        if (descriptor && typeof key === "string") {
            if (descriptor === "json") {
                target._raw.set(key, value);
            }
            else if ("encoded" in descriptor) {
                target._raw.set(key, encodeSync(descriptor.encoded)(value));
            }
            else if (isRefEncoded(descriptor)) {
                target._raw.set(key, value.id);
                subscriptionsScopes.get(target)?.onRefAccessedOrSet(value.id);
            }
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, attributes) {
        var _b;
        if ("value" in attributes &&
            typeof attributes.value === "object" &&
            SchemaInit in attributes.value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[key] =
                attributes.value[SchemaInit];
            tryInit(target);
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, attributes);
        }
    },
    ownKeys(target) {
        const keys = Reflect.ownKeys(target).filter((k) => k !== ItemsSym);
        // for (const key of Reflect.ownKeys(target._schema)) {
        //     if (key !== ItemsSym && !keys.includes(key)) {
        //         keys.push(key);
        //     }
        // }
        for (const key of target._raw.keys()) {
            if (!keys.includes(key)) {
                keys.push(key);
            }
        }
        return keys;
    },
    getOwnPropertyDescriptor(target, key) {
        if (key in target) {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
        else {
            const descriptor = (target._schema[key] ||
                target._schema[ItemsSym]);
            if (descriptor || key in target._raw.ops) {
                return {
                    enumerable: true,
                    configurable: true,
                    writable: true,
                };
            }
        }
    },
    deleteProperty(target, key) {
        const descriptor = (target._schema[key] ||
            target._schema[ItemsSym]);
        if (typeof key === "string" && descriptor) {
            target._raw.delete(key);
            return true;
        }
        else {
            return Reflect.deleteProperty(target, key);
        }
    },
};
//# sourceMappingURL=coMap.js.map