var _a;
import { cojsonInternals } from "cojson";
import { ItemsSym, Account, CoValueBase, Ref, inspect, co, InitValues, SchemaInit, isRefEncoded, } from "../internal.js";
import { encodeSync, decodeSync } from "@effect/schema/Schema";
export class CoStream extends CoValueBase {
    static Of(item) {
        var _b, _c;
        return _c = class CoStreamOf extends _a {
                constructor() {
                    super(...arguments);
                    this[_b] = item;
                }
            },
            _b = co.items,
            _c;
    }
    get _schema() {
        return this.constructor._schema;
    }
    get byMe() {
        return this[this._loadedAs.id];
    }
    get inCurrentSession() {
        return this.perSession[this._loadedAs.sessionID];
    }
    constructor(init, options) {
        super();
        if ("fromRaw" in options) {
            Object.defineProperties(this, {
                id: {
                    value: options.fromRaw.id,
                    enumerable: false,
                },
                _raw: { value: options.fromRaw, enumerable: false },
            });
        }
        else {
            this[InitValues] = {
                init,
                owner: options.owner,
            };
        }
        return new Proxy(this, CoStreamProxyHandler);
    }
    push(...items) {
        for (const item of items) {
            this.pushItem(item);
        }
    }
    pushItem(item) {
        const itemDescriptor = this._schema[ItemsSym];
        if (itemDescriptor === "json") {
            this._raw.push(item);
        }
        else if ("encoded" in itemDescriptor) {
            this._raw.push(encodeSync(itemDescriptor.encoded)(item));
        }
        else if (isRefEncoded(itemDescriptor)) {
            this._raw.push(item.id);
        }
    }
    toJSON() {
        const itemDescriptor = this._schema[ItemsSym];
        const mapper = itemDescriptor === "json"
            ? (v) => v
            : "encoded" in itemDescriptor
                ? encodeSync(itemDescriptor.encoded)
                : (v) => v && v.id;
        return {
            id: this.id,
            _type: this._type,
            ...Object.fromEntries(Object.entries(this).map(([account, entry]) => [
                account,
                mapper(entry.value),
            ])),
            in: Object.fromEntries(Object.entries(this.perSession).map(([session, entry]) => [
                session,
                mapper(entry.value),
            ])),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    static schema(def) {
        this._schema || (this._schema = {});
        Object.assign(this._schema, def);
    }
}
_a = CoStream;
(() => {
    _a.prototype._type = "CoStream";
})();
function entryFromRawEntry(accessFrom, rawEntry, loadedAs, accountID, itemField) {
    return {
        get value() {
            if (itemField === "json") {
                return rawEntry.value;
            }
            else if ("encoded" in itemField) {
                return decodeSync(itemField.encoded)(rawEntry.value);
            }
            else if (isRefEncoded(itemField)) {
                return this.ref?.accessFrom(accessFrom);
            }
            else {
                throw new Error("Invalid item field schema");
            }
        },
        get ref() {
            if (itemField !== "json" && isRefEncoded(itemField)) {
                const rawId = rawEntry.value;
                return new Ref(rawId, loadedAs, itemField);
            }
            else {
                return undefined;
            }
        },
        get by() {
            return (accountID &&
                new Ref(accountID, loadedAs, Account)?.accessFrom(accessFrom));
        },
        madeAt: rawEntry.at,
        tx: rawEntry.tx,
    };
}
function init(stream) {
    const init = stream[InitValues];
    if (!init)
        return;
    const raw = init.owner._raw.createStream();
    Object.defineProperties(stream, {
        id: {
            value: raw.id,
            enumerable: false,
        },
        _raw: { value: raw, enumerable: false },
    });
    if (init.init) {
        stream.push(...init.init);
    }
    delete stream[InitValues];
}
export const CoStreamProxyHandler = {
    get(target, key, receiver) {
        if (typeof key === "string" && key.startsWith("co_")) {
            const rawEntry = target._raw.lastItemBy(key);
            if (!rawEntry)
                return;
            const entry = entryFromRawEntry(receiver, rawEntry, target._loadedAs, key, target._schema[ItemsSym]);
            Object.defineProperty(entry, "all", {
                get: () => {
                    const allRawEntries = target._raw.itemsBy(key);
                    return (function* () {
                        while (true) {
                            const rawEntry = allRawEntries.next();
                            if (rawEntry.done)
                                return;
                            yield entryFromRawEntry(receiver, rawEntry.value, target._loadedAs, key, target._schema[ItemsSym]);
                        }
                    })();
                },
            });
            return entry;
        }
        else if (key === "perSession") {
            return new Proxy({}, CoStreamPerSessionProxyHandler(target, receiver));
        }
        else {
            return Reflect.get(target, key, receiver);
        }
    },
    set(target, key, value, receiver) {
        var _b;
        if (key === ItemsSym &&
            typeof value === "object" &&
            SchemaInit in value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[ItemsSym] =
                value[SchemaInit];
            init(target);
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, descriptor) {
        var _b;
        if (descriptor.value &&
            key === ItemsSym &&
            typeof descriptor.value === "object" &&
            SchemaInit in descriptor.value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[ItemsSym] =
                descriptor.value[SchemaInit];
            init(target);
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, descriptor);
        }
    },
    ownKeys(target) {
        const keys = Reflect.ownKeys(target);
        for (const accountID of target._raw.accounts()) {
            keys.push(accountID);
        }
        return keys;
    },
    getOwnPropertyDescriptor(target, key) {
        if (typeof key === "string" && key.startsWith("co_")) {
            return {
                configurable: true,
                enumerable: true,
                writable: false,
            };
        }
        else {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
    },
};
const CoStreamPerSessionProxyHandler = (innerTarget, accessFrom) => ({
    get(_target, key, receiver) {
        if (typeof key === "string" && key.includes("session")) {
            const sessionID = key;
            const rawEntry = innerTarget._raw.lastItemIn(sessionID);
            if (!rawEntry)
                return;
            const by = cojsonInternals.accountOrAgentIDfromSessionID(sessionID);
            const entry = entryFromRawEntry(accessFrom, rawEntry, innerTarget._loadedAs, cojsonInternals.isAccountID(by)
                ? by
                : undefined, innerTarget._schema[ItemsSym]);
            Object.defineProperty(entry, "all", {
                get: () => {
                    const allRawEntries = innerTarget._raw.itemsIn(sessionID);
                    return (function* () {
                        while (true) {
                            const rawEntry = allRawEntries.next();
                            if (rawEntry.done)
                                return;
                            yield entryFromRawEntry(accessFrom, rawEntry.value, innerTarget._loadedAs, cojsonInternals.isAccountID(by)
                                ? by
                                : undefined, innerTarget._schema[ItemsSym]);
                        }
                    })();
                },
            });
            return entry;
        }
        else {
            return Reflect.get(innerTarget, key, receiver);
        }
    },
    ownKeys() {
        return innerTarget._raw.sessions();
    },
    getOwnPropertyDescriptor(target, key) {
        if (typeof key === "string" && key.startsWith("co_")) {
            return {
                configurable: true,
                enumerable: true,
                writable: false,
            };
        }
        else {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
    },
});
export class BinaryCoStream extends CoValueBase {
    constructor(_init, options) {
        super();
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            raw = rawOwner.createBinaryStream();
        }
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
    }
    getChunks(options) {
        return this._raw.getBinaryChunks(options?.allowUnfinished);
    }
    start(options) {
        this._raw.startBinaryStream(options);
    }
    push(data) {
        this._raw.pushBinaryStreamChunk(data);
    }
    end() {
        this._raw.endBinaryStream();
    }
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
            ...this.getChunks(),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
}
//# sourceMappingURL=coStream.js.map