import { Effect, Sink, Stream } from "effect";
import { RawAccount } from "cojson";
import { Account, AccountCtx, Group, SubscriptionScope, Ref, inspect, } from "../internal.js";
export function isCoValue(value) {
    return value && value._type !== undefined;
}
export function isCoValueClass(value) {
    return typeof value === "function" && value.fromRaw !== undefined;
}
export class CoValueBase {
    get _owner() {
        return this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
    }
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    constructor(..._args) { }
    static fromRaw(raw) {
        return new this(undefined, { fromRaw: raw });
    }
    static loadEf(id) {
        return Effect.gen(this, function* (_) {
            const account = yield* _(AccountCtx);
            return yield* _(new Ref(id, account, this).loadEf());
        });
    }
    static load(id, options) {
        return new Ref(id, options.as, this).load(options?.onProgress && { onProgress: options.onProgress });
    }
    static subscribe(id, options, onUpdate) {
        void Effect.runPromise(Effect.provideService(this.subscribeEf(id).pipe(Stream.run(Sink.forEach((update) => Effect.sync(() => onUpdate(update))))), AccountCtx, options.as));
        return function unsubscribe() { };
    }
    static subscribeEf(id) {
        return Stream.fromEffect(this.loadEf(id)).pipe(Stream.flatMap((value) => Stream.asyncScoped((emit) => Effect.gen(this, function* (_) {
            const subscription = new SubscriptionScope(value, this, (update) => {
                void emit.single(update);
            });
            yield* _(Effect.addFinalizer(() => Effect.sync(() => subscription.unsubscribeAll())));
        }))));
    }
    subscribe(listener) {
        return this.constructor.subscribe(this.id, { as: this._loadedAs }, listener);
    }
    subscribeEf() {
        return Stream.provideService(this.constructor.subscribeEf(this.id), AccountCtx, this._loadedAs);
    }
    toJSON() {
        return {
            id: this.id,
            type: this._type,
            error: "unknown CoValue class",
        };
    }
    [inspect]() {
        return this.toJSON();
    }
}
//# sourceMappingURL=interfaces.js.map