export const subscriptionsScopes = new WeakMap();
export class SubscriptionScope {
    constructor(root, rootSchema, onUpdate) {
        this.scopeID = `scope-${Math.random().toString(36).slice(2)}`;
        this.entries = new Map();
        this.scheduledUpdate = false;
        this.rootEntry = {
            state: "loaded",
            value: root,
            rawUnsub: () => { }, // placeholder
        };
        this.entries.set(root.id, this.rootEntry);
        subscriptionsScopes.set(root, this);
        this.subscriber = root._loadedAs;
        this.onUpdate = onUpdate;
        this.rootEntry.rawUnsub = root._raw.core.subscribe((rawUpdate) => {
            if (!rawUpdate)
                return;
            this.rootEntry.value = rootSchema.fromRaw(rawUpdate);
            // console.log("root update", this.rootEntry.value.toJSON());
            subscriptionsScopes.set(this.rootEntry.value, this);
            this.scheduleUpdate();
        });
    }
    scheduleUpdate() {
        if (!this.scheduledUpdate) {
            this.scheduledUpdate = true;
            queueMicrotask(() => {
                this.scheduledUpdate = false;
                this.onUpdate(this.rootEntry.value);
            });
        }
    }
    onRefAccessedOrSet(accessedOrSetId) {
        // console.log("onRefAccessedOrSet", this.scopeID, accessedOrSetId);
        if (!accessedOrSetId) {
            return;
        }
        if (!this.entries.has(accessedOrSetId)) {
            const loadingEntry = {
                state: "loading",
                immediatelyUnsub: false,
            };
            this.entries.set(accessedOrSetId, loadingEntry);
            void this.subscriber._raw.core.node
                .loadCoValueCore(accessedOrSetId)
                .then((core) => {
                if (loadingEntry.state === "loading" &&
                    loadingEntry.immediatelyUnsub) {
                    return;
                }
                if (core !== "unavailable") {
                    const entry = {
                        state: "loaded",
                        rawUnsub: () => { }, // placeholder
                    };
                    this.entries.set(accessedOrSetId, entry);
                    const rawUnsub = core.subscribe((rawUpdate) => {
                        // console.log("ref update", this.scopeID, accessedOrSetId, JSON.stringify(rawUpdate))
                        if (!rawUpdate)
                            return;
                        this.scheduleUpdate();
                    });
                    entry.rawUnsub = rawUnsub;
                }
            });
        }
    }
    unsubscribeAll() {
        for (const entry of this.entries.values()) {
            if (entry.state === "loaded") {
                entry.rawUnsub();
            }
            else {
                entry.immediatelyUnsub = true;
            }
        }
        this.entries.clear();
    }
}
//# sourceMappingURL=subscriptionScope.js.map