import { expect, describe, test, beforeEach } from "vitest";
import { webcrypto } from "node:crypto";
import { Account, jazzReady, CoMap, co, Group } from "..";
if (!("crypto" in globalThis)) {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    globalThis.crypto = webcrypto;
}
beforeEach(async () => {
    await jazzReady;
});
describe("Custom accounts and groups", async () => {
    var _a;
    class CustomProfile extends CoMap {
        constructor() {
            super(...arguments);
            this.name = co.string;
            this.color = co.string;
        }
    }
    class CustomAccount extends Account {
        constructor() {
            super(...arguments);
            this.profile = co.ref(CustomProfile);
            this.root = co.ref(CoMap);
        }
        migrate(creationProps) {
            if (creationProps) {
                const profileGroup = new Group({ owner: this });
                profileGroup.addMember("everyone", "reader");
                this.profile = new CustomProfile({ name: creationProps.name, color: "blue" }, { owner: this });
            }
        }
    }
    class CustomGroup extends Group {
        constructor() {
            super(...arguments);
            this.profile = co.null;
            this.root = co.null;
            this[_a] = co.ref(CustomAccount);
        }
    }
    _a = co.members;
    test("Custom account and group", async () => {
        const me = await CustomAccount.create({
            creationProps: { name: "Hermes Puggington" },
        });
        expect(me.profile).toBeDefined();
        expect(me.profile?.name).toBe("Hermes Puggington");
        expect(me.profile?.color).toBe("blue");
        const group = new CustomGroup({ owner: me });
        group.addMember("everyone", "reader");
        expect(group.members).toMatchObject([
            { id: me.id, role: "admin" },
            { id: "everyone", role: "reader" },
        ]);
        await new Promise((resolve) => {
            group.subscribe((update) => {
                const myProfile = update.members.find((member) => {
                    return member.id === me.id && member.account?.profile;
                });
                if (myProfile) {
                    expect(myProfile.account?.profile?.name).toBe("Hermes Puggington");
                    expect(myProfile.account?.profile?.color).toBe("blue");
                    resolve();
                }
            });
        });
    });
});
//# sourceMappingURL=groupsAndAccounts.test.js.map